import platform
import unittest

from pathlib import Path

try:
    from unittest.mock import patch
except ImportError:
    # Python 2
    from mock import patch

import numpy as np

from PIL import Image

try:
    import cv2
except ImportError:
    cv2 = None

try:
    import imageio
except ImportError:
    imageio = None

from pyzbar import wrapper
from pyzbar.pyzbar import (
    decode, Decoded, Rect, ZBarSymbol, EXTERNAL_DEPENDENCIES, ORIENTATION_AVAILABLE
)
from pyzbar.pyzbar_error import PyZbarError


TESTDATA = Path(__file__).parent


class TestDecode(unittest.TestCase):
    EXPECTED_CODE128 = [
        Decoded(
            data=b'Foramenifera',
            type='CODE128',
            rect=Rect(left=37, top=550, width=324, height=76),
            polygon=[(37, 551), (37, 625), (361, 626), (361, 550)],
            orientation='UP' if ORIENTATION_AVAILABLE else None,
            quality=77,
        ),
        Decoded(
            data=b'Rana temporaria',
            type='CODE128',
            rect=Rect(left=4, top=0, width=390, height=76),
            polygon=[(4, 1), (4, 75), (394, 76), (394, 0)],
            orientation='UP' if ORIENTATION_AVAILABLE else None,
            quality=77,
        ),
    ]

    EXPECTED_CODE128_NULL_CHARACTER = [
        Decoded(
            data=b'Hello\0Goodbye',
            type='CODE128',
            rect=Rect(left=4, top=0, width=390, height=75),
            polygon=[(4, 1), (4, 75), (394, 74), (394, 0)],
            orientation="UP" if ORIENTATION_AVAILABLE else None,
            quality=76,
        ),
    ]

    EXPECTED_QRCODE = [
        Decoded(
            b'Thalassiodracon',
            type='QRCODE',
            rect=Rect(left=27, top=27, width=145, height=145),
            polygon=[(27, 27), (27, 172), (172, 172), (172, 27)],
            orientation='UP' if ORIENTATION_AVAILABLE else None,
            quality=1,
        ),
    ]

    # Two barcodes, both with same content
    EXPECTED_QRCODE_ROTATED = [
        Decoded(
            data=b'Thalassiodracon',
            type='QRCODE',
            rect=Rect(left=173, top=10, width=205, height=205),
            polygon=[(173, 113), (276, 215), (378, 113), (276, 10)],
            orientation='UP' if ORIENTATION_AVAILABLE else None,
            quality=1,
        ),
        Decoded(
            data=b'Thalassiodracon',
            type='QRCODE',
            rect=Rect(left=32, top=208, width=158, height=158),
            polygon=[(32, 352), (177, 366), (190, 222), (46, 208)],
            orientation='RIGHT' if ORIENTATION_AVAILABLE else None,
            quality=1,
        ),
    ]

    @classmethod
    def setUpClass(cls):
        cls.code128, cls.code128_null_character, cls.qrcode, cls.qrcode_rotated, cls.empty = (
            Image.open(str(TESTDATA.joinpath(fname)))
            for fname in
            ('code128.png', 'code128_null_character.png', 'qrcode.png', 'qrcode_rotated.png', 'empty.png')
        )

        cls.maxDiff = None

        # assertRaisesRegexp was a deprecated alias removed in Python 3.11
        if not hasattr(cls, 'assertRaisesRegex'):
            cls.assertRaisesRegex = cls.assertRaisesRegexp

    @classmethod
    def tearDownClass(cls):
        cls.code128 = cls.code128_null_character = cls.qrcode = cls.qrcode_rotated = cls.empty = None

    def test_decode_code128(self):
        "Read both barcodes in `code128.png`"
        res = decode(self.code128)
        self.assertEqual(self.EXPECTED_CODE128, res)

    def test_decode_code128_null_character(self):
        "Read barcode in `code128_null_character.png` containing a null character"
        res = decode(self.code128_null_character)
        self.assertEqual(self.EXPECTED_CODE128_NULL_CHARACTER, res)

    def test_decode_qrcode(self):
        "Read barcode in `qrcode.png`"
        res = decode(self.qrcode)
        self.assertEqual(self.EXPECTED_QRCODE, res)

    def test_decode_qrcode_rotated(self):
        "Read barcode in `qrcode_rotated.png`"
        # Test computation of the polygon around the barcode
        res = decode(self.qrcode_rotated)
        self.assertEqual(self.EXPECTED_QRCODE_ROTATED, res)

    def test_symbols(self):
        "Read only qrcodes in `qrcode.png`"
        res = decode(self.qrcode, symbols=[ZBarSymbol.QRCODE])
        self.assertEqual(self.EXPECTED_QRCODE, res)

    def test_symbols_not_present(self):
        "Read only code128 in `qrcode.png`"
        res = decode(self.qrcode, symbols=[ZBarSymbol.CODE128])
        self.assertEqual([], res)

    def test_decode_tuple(self):
        "Read barcodes in pixels"
        pixels = self.code128.copy().convert('L').tobytes()
        width, height = self.code128.size
        res = decode((pixels, width, height))
        self.assertEqual(self.EXPECTED_CODE128, res)

    def test_unsupported_bpp(self):
        pixels = self.code128.tobytes()
        width, height = self.code128.size
        self.assertRaises(PyZbarError, decode, (pixels, width, height))

    def test_empty(self):
        "Do not show any output for an image that does not contain a barcode"
        res = decode(self.empty)
        self.assertEqual([], res)

    def test_decode_numpy(self):
        "Read image using Pillow and convert to numpy.ndarray"
        res = decode(np.asarray(self.code128))
        self.assertEqual(self.EXPECTED_CODE128, res)

    @unittest.skipIf(imageio is None, 'imageio not installed')
    def test_decode_imageio(self):
        "Read image using imageio"
        res = decode(imageio.imread(TESTDATA.joinpath('code128.png')))
        self.assertEqual(self.EXPECTED_CODE128, res)

    @unittest.skipIf(cv2 is None, 'OpenCV not installed')
    def test_decode_opencv(self):
        "Read image using OpenCV"
        res = decode(cv2.imread(str(TESTDATA.joinpath('code128.png'))))
        self.assertEqual(self.EXPECTED_CODE128, res)

    @patch('pyzbar.pyzbar.zbar_image_first_symbol', autospec=True)
    def test_unrecognised_symbol_type(self, zbar_image_first_symbol):
        "The type of the first symbol is not recognised"
        def zbar_image_first_symbol_set_symbol_type(image):
            symbol = wrapper.zbar_image_first_symbol(image)
            if symbol:
                symbol.contents.type = -1
            return symbol

        zbar_image_first_symbol.side_effect = zbar_image_first_symbol_set_symbol_type

        res = decode(np.asarray(self.code128))

        expected = [
            self.EXPECTED_CODE128[0]._replace(type='Unrecognised type [-1]')
        ] + self.EXPECTED_CODE128[1:]
        self.assertEqual(expected, res)

    def test_external_dependencies(self):
        "External dependencies"
        if 'Windows' == platform.system():
            self.assertEqual(2, len(EXTERNAL_DEPENDENCIES))
            self.assertTrue(
                any('libiconv' in d._name for d in EXTERNAL_DEPENDENCIES)
            )
        self.assertTrue(
            any('libzbar' in d._name for d in EXTERNAL_DEPENDENCIES)
        )

    @patch('pyzbar.pyzbar.zbar_image_create', autospec=True)
    def test_zbar_image_create_fail(self, zbar_image_create):
        zbar_image_create.return_value = None
        self.assertRaisesRegex(
            PyZbarError, 'Could not create zbar image', decode, self.code128
        )
        zbar_image_create.assert_called_once_with()

    @patch('pyzbar.pyzbar.zbar_image_scanner_create', autospec=True)
    def test_zbar_image_scanner_create_fail(self, zbar_image_scanner_create):
        zbar_image_scanner_create.return_value = None
        self.assertRaisesRegex(
            PyZbarError, 'Could not create image scanner', decode, self.code128
        )
        zbar_image_scanner_create.assert_called_once_with()

    @patch('pyzbar.pyzbar.zbar_scan_image', autospec=True)
    def test_zbar_scan_image_fail(self, zbar_scan_image):
        zbar_scan_image.return_value = -1
        self.assertRaisesRegex(
            PyZbarError, 'Unsupported image format', decode, self.code128
        )
        self.assertEqual(1, zbar_scan_image.call_count)

    def test_unsupported_bits_per_pixel(self):
        # 16 bits-per-pixel
        data = (list(range(3 * 3 * 2)), 3, 3)
        self.assertRaisesRegex(
            PyZbarError,
            r'Unsupported bits-per-pixel \[16\]. Only \[8\] is supported.',
            decode, data
        )
        self.assertRaises(PyZbarError, decode, data)

    def test_inconsistent_dimensions(self):
        # Ten bytes but width x height indicates nine bytes
        data = (list(range(10)), 3, 3)
        self.assertRaisesRegex(
            PyZbarError,
            (
                r'Inconsistent dimensions: image data of 10 bytes is not '
                r'divisible by \(width x height = 9\)'
            ),
            decode, data
        )


if __name__ == '__main__':
    unittest.main()



from collections import namedtuple
from contextlib import contextmanager
from ctypes import cast, c_void_p, string_at

from .locations import bounding_box, convex_hull, Point, Rect
from .pyzbar_error import PyZbarError
from .wrapper import (
    zbar_image_scanner_set_config,
    zbar_image_scanner_create, zbar_image_scanner_destroy,
    zbar_image_create, zbar_image_destroy, zbar_image_set_format,
    zbar_image_set_size, zbar_image_set_data, zbar_scan_image,
    zbar_image_first_symbol, zbar_symbol_get_data_length,
    zbar_symbol_get_data, zbar_symbol_get_orientation,
    zbar_symbol_get_loc_size, zbar_symbol_get_loc_x, zbar_symbol_get_loc_y,
    zbar_symbol_get_quality, zbar_symbol_next, ZBarConfig, ZBarOrientation,
    ZBarSymbol, EXTERNAL_DEPENDENCIES,
)

__all__ = [
    'decode', 'Point', 'Rect', 'Decoded', 'ZBarSymbol', 'EXTERNAL_DEPENDENCIES', 'ORIENTATION_AVAILABLE'
]


ORIENTATION_AVAILABLE = zbar_symbol_get_orientation is not None

Decoded = namedtuple('Decoded', 'data type rect polygon quality orientation')

# ZBar's magic 'fourcc' numbers that represent image formats
_FOURCC = {
    'L800': 808466521,
    'GRAY': 1497715271
}

_RANGEFN = getattr(globals(), 'xrange', range)


@contextmanager
def _image():
    """A context manager for `zbar_image`, created and destoyed by
    `zbar_image_create` and `zbar_image_destroy`.

    Yields:
        POINTER(zbar_image): The created image

    Raises:
        PyZbarError: If the image could not be created.
    """
    image = zbar_image_create()
    if not image:
        raise PyZbarError('Could not create zbar image')
    else:
        try:
            yield image
        finally:
            zbar_image_destroy(image)


@contextmanager
def _image_scanner():
    """A context manager for `zbar_image_scanner`, created and destroyed by
    `zbar_image_scanner_create` and `zbar_image_scanner_destroy`.

    Yields:
        POINTER(zbar_image_scanner): The created scanner

    Raises:
        PyZbarError: If the decoder could not be created.
    """
    scanner = zbar_image_scanner_create()
    if not scanner:
        raise PyZbarError('Could not create image scanner')
    else:
        try:
            yield scanner
        finally:
            zbar_image_scanner_destroy(scanner)


def _symbols_for_image(image):
    """Generator of symbols.

    Args:
        image: `zbar_image`

    Yields:
        POINTER(zbar_symbol): Symbol
    """
    symbol = zbar_image_first_symbol(image)
    while symbol:
        yield symbol
        symbol = zbar_symbol_next(symbol)


def _decode_symbols(symbols):
    """Generator of decoded symbol information.

    Args:
        symbols: iterable of instances of `POINTER(zbar_symbol)`

    Yields:
        Decoded: decoded symbol
    """
    for symbol in symbols:
        data = string_at(
            zbar_symbol_get_data(symbol),
            zbar_symbol_get_data_length(symbol)
        )
        # The 'type' int should be a value in the ZBarSymbol enumeration
        try:
            symbol_type = ZBarSymbol(symbol.contents.type)
        except ValueError:
            # This release of zbar supports a type that pyzbar does not know about
            symbol_type = "Unrecognised type [{0}]".format(symbol.contents.type)
        else:
            symbol_type = symbol_type.name

        quality = zbar_symbol_get_quality(symbol)
        polygon = convex_hull(
            (
                zbar_symbol_get_loc_x(symbol, index),
                zbar_symbol_get_loc_y(symbol, index)
            )
            for index in _RANGEFN(zbar_symbol_get_loc_size(symbol))
        )

        if zbar_symbol_get_orientation:
            orientation = ZBarOrientation(zbar_symbol_get_orientation(symbol)).name
        else:
            orientation = None

        yield Decoded(
            data=data,
            type=symbol_type,
            rect=bounding_box(polygon),
            polygon=polygon,
            orientation=orientation,
            quality=quality,
        )


def _pixel_data(image):
    """Returns (pixels, width, height)

    Returns:
        :obj: `tuple` (pixels, width, height)
    """
    # Test for PIL.Image, numpy.ndarray, and imageio.core.util without
    # requiring that cv2, PIL, or imageio are installed.

    image_type = str(type(image))
    if 'PIL.' in image_type:
        if 'L' != image.mode:
            image = image.convert('L')
        pixels = image.tobytes()
        width, height = image.size
    elif 'numpy.ndarray' in image_type or 'imageio.core.util' in image_type:
        # Different versions of imageio use a subclass of numpy.ndarray
        # called either imageio.core.util.Image or imageio.core.util.Array.
        if 3 == len(image.shape):
            # Take just the first channel
            image = image[:, :, 0]
        if 'uint8' != str(image.dtype):
            image = image.astype('uint8')
        try:
            pixels = image.tobytes()
        except AttributeError:
            # `numpy.ndarray.tobytes()` introduced in `numpy` 1.9.0 - use the
            # older `tostring` method.
            pixels = image.tostring()
        height, width = image.shape[:2]
    else:
        # image should be a tuple (pixels, width, height)
        pixels, width, height = image

        # Check dimensions
        if 0 != len(pixels) % (width * height):
            raise PyZbarError(
                (
                    'Inconsistent dimensions: image data of {0} bytes is not '
                    'divisible by (width x height = {1})'
                ).format(len(pixels), (width * height))
            )

    # Compute bits-per-pixel
    bpp = 8 * len(pixels) // (width * height)
    if 8 != bpp:
        raise PyZbarError(
            'Unsupported bits-per-pixel [{0}]. Only [8] is supported.'.format(
                bpp
            )
        )

    return pixels, width, height


def decode(image, symbols=None):
    """Decodes datamatrix barcodes in `image`.

    Args:
        image: `numpy.ndarray`, `PIL.Image` or tuple (pixels, width, height)
        symbols: iter(ZBarSymbol) the symbol types to decode; if `None`, uses
            `zbar`'s default behaviour, which is to decode all symbol types.

    Returns:
        :obj:`list` of :obj:`Decoded`: The values decoded from barcodes.
    """
    pixels, width, height = _pixel_data(image)

    results = []
    with _image_scanner() as scanner:
        if symbols:
            # Disable all but the symbols of interest
            disable = set(ZBarSymbol).difference(symbols)
            for symbol in disable:
                zbar_image_scanner_set_config(
                    scanner, symbol, ZBarConfig.CFG_ENABLE, 0
                )
            # I think it likely that zbar will detect all symbol types by
            # default, in which case enabling the types of interest is
            # redundant but it seems sensible to be over-cautious and enable
            # them.
            for symbol in symbols:
                zbar_image_scanner_set_config(
                    scanner, symbol, ZBarConfig.CFG_ENABLE, 1
                )
        with _image() as img:
            zbar_image_set_format(img, _FOURCC['L800'])
            zbar_image_set_size(img, width, height)
            zbar_image_set_data(img, cast(pixels, c_void_p), len(pixels), None)
            decoded = zbar_scan_image(scanner, img)
            if decoded < 0:
                raise PyZbarError('Unsupported image format')
            else:
                results.extend(_decode_symbols(_symbols_for_image(img)))

    return results



#!/usr/bin/env python
import sys

import pyzbar


SCRIPTS = ['read_zbar']

# Optional dependency
PILLOW = 'Pillow>=3.2.0'

URL = 'https://github.com/NaturalHistoryMuseum/pyzbar/'


def readme():
    # TODO IOError on Python 2.x. FileNotFoundError on Python 3.x.
    try:
        with open('README.rst') as f:
            return f.read()
    except:
        return 'Visit {0} for more details.'.format(URL)


setup_data = {
    'name': 'pyzbar',
    'version': pyzbar.__version__,
    'author': 'Lawrence Hudson',
    'author_email': 'quicklizard@googlemail.com',
    'url': URL,
    'license': 'MIT',
    'description': pyzbar.__doc__,
    'long_description': readme(),
    'long_description_content_type': 'text/x-rst',
    'packages': ['pyzbar', 'pyzbar.scripts', 'pyzbar.tests'],
    'test_suite': 'pyzbar.tests',
    'scripts': ['pyzbar/scripts/{0}.py'.format(script) for script in SCRIPTS],
    'entry_points': {
        'console_scripts': [
            '{0}=pyzbar.scripts.{0}:main'.format(script) for script in SCRIPTS
        ],
    },
    'extras_require': {
        ':python_version=="2.7"': ['enum34>=1.1.6', 'pathlib>=1.0.1'],
        'scripts': [
            PILLOW,
        ],
    },
    'tests_require': [
        # TODO How to specify OpenCV? 'cv2>=2.4.8',
        'mock>=2.0.0; python_version=="2.7"',
        'numpy>=1.8.2',
        PILLOW,
    ],
    'include_package_data': True,
    'classifiers': [
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Topic :: Utilities',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
}


def setuptools_setup():
    from setuptools import setup
    setup(**setup_data)


if (2, 7) == sys.version_info[:2] or (3, 5) <= sys.version_info:
    setuptools_setup()
else:
    sys.exit('Python versions 2.7 and >= 3.5 are supported')



from collections import namedtuple
from itertools import chain
from operator import itemgetter


__all__ = ['bounding_box', 'convex_hull', 'Point', 'Rect']


Point = namedtuple('Point', ['x', 'y'])
Rect = namedtuple('Rect', ['left', 'top', 'width', 'height'])


def bounding_box(locations):
    """Computes the bounding box of an iterable of (x, y) coordinates.

    Args:
        locations: iterable of (x, y) tuples.

    Returns:
        `Rect`: Coordinates of the bounding box.
    """
    x_values = list(map(itemgetter(0), locations))
    x_min, x_max = min(x_values), max(x_values)
    y_values = list(map(itemgetter(1), locations))
    y_min, y_max = min(y_values), max(y_values)
    return Rect(x_min, y_min, x_max - x_min, y_max - y_min)


def convex_hull(points):
    """Computes the convex hull of an iterable of (x, y) coordinates.

    Args:
        points: iterable of (x, y) tuples.

    Returns:
        `list`: instances of `Point` - vertices of the convex hull in
        counter-clockwise order, starting from the vertex with the
        lexicographically smallest coordinates.

    Andrew's monotone chain algorithm. O(n log n) complexity.
    https://en.wikibooks.org/wiki/Algorithm_Implementation/Geometry/Convex_hull/Monotone_chain
    """

    def is_not_clockwise(p0, p1, p2):
        return 0 <= (
            (p1[0] - p0[0]) * (p2[1] - p0[1]) -
            (p1[1] - p0[1]) * (p2[0] - p0[0])
        )

    def go(points_):
        res = []
        for p in points_:
            while 1 < len(res) and is_not_clockwise(res[-2], res[-1], p):
                res.pop()
            res.append(p)

        # The last point in each list is the first point in the other list
        res.pop()

        return res

    # Discard duplicates and sort by x then y
    points = sorted(set(points))

    # Algorithm needs at least two points
    hull = (
        points if len(points) < 2 else chain(go(points), go(reversed(points)))
    )

    return list(map(Point._make, hull))
